@@ -174,12 +174,16 @@ static struct block *block_new_insert(char *name, int start, int end)
 		dest->next=parent2;
 
 		parent2->prev=dest;
+		if(parent2->next != NULL) {
+			parent2->next->prev = parent2;
+		}
 	}
 
 	return(dest);
 }
 
-struct block *block_new(char *name, int start, int end, enum blocktype type)
+struct block *block_new(char *name, int start, int end, enum blocktype type,
+			enum labeltype start_label, enum labeltype end_label)
 {
 	struct block *dest;
 	struct symbol *symb;
@@ -191,18 +195,33 @@ struct block *block_new(char *name, int start, int end, enum blocktype type)
 	dest=block_new_insert(name, start, end);
 	if(dest==NULL) return NULL;
 
-	snprintf(tmp, MAX_LINE_LEN, "%s_start", name);
-	symb=symbol_new(tmp, start, 100, 0);
+	if(start_label == LABEL_START_END || start_label == LABEL_FIRST_LAST) {
 
-	snprintf(tmp, MAX_LINE_LEN, "\n; BLOCK '%s' (start 0x%04x "
+		if(start_label == LABEL_START_END) {
+			snprintf(tmp, MAX_LINE_LEN, "%s_start", name);
+		} else {
+			snprintf(tmp, MAX_LINE_LEN, "%s_first", name);
+		}
+
+		symb = symbol_new(tmp, start, 100, 0);
+
+		snprintf(tmp, MAX_LINE_LEN, "\n; BLOCK '%s' (start 0x%04x "
 					"end 0x%04x)\n", name, start, end);
 
-	symb->comment=strdup(tmp);
+		symb->comment = strdup(tmp);
+	}
 
-	snprintf(tmp, MAX_LINE_LEN, "%s_end", name);
-	symbol_new(tmp, end, 0, 0);
+	if(end_label == LABEL_START_END) {
+		snprintf(tmp, MAX_LINE_LEN, "%s_end", name);
+		symbol_new(tmp, end, 0, 0);
+	} else if(end_label == LABEL_FIRST_LAST) {
+		snprintf(tmp, MAX_LINE_LEN, "%s_last", name);
+		symbol_new(tmp, end - 1, 0, 0);
+	}
 
 	dest->type=type;
+	dest->start_label = start_label;
+	dest->end_label = end_label;
 
 	return dest;
 }
@@ -215,6 +234,7 @@ int block_load_file(char *filename)
 	int r;
 	int start,end;
 	enum blocktype type;
+	enum labeltype start_label, end_label;
 	int lineno;
 	char *name;
 
@@ -229,7 +249,8 @@ int block_load_file(char *filename)
 
 	lineno=1;
 	while(fgets(line, MAX_LINE_LEN, f)!=NULL) {
-		r=block_load_line(line, &name, &start, &end, &type);
+		r=block_load_line(line, &name, &start, &end, &type,
+						&start_label, &end_label);
 		if(r<0) {
 			msg(0, "       at line %d of '%s'\n", lineno, filename);
 			fclose(f);
@@ -237,7 +258,8 @@ int block_load_file(char *filename)
 		}
 
 		if(r==1) {
-			blk=block_new(name, start, end, type);
+			blk=block_new(name, start, end, type,
+						start_label, end_label);
 			if(blk==NULL) {
 				msg(0, "       at line %d of '%s'\n", lineno, 
 								filename);
@@ -256,7 +278,9 @@ int block_load_file(char *filename)
 
 /* TODO: Use split_line() from utils.c */
 int block_load_line(char *line, char **name, 
-				int *start, int *end, enum blocktype *type)
+				int *start, int *end, enum blocktype *type,
+				enum labeltype *start_label,
+				enum labeltype *end_label)
 {
 	int r;
 
@@ -264,6 +288,7 @@ int block_load_line(char *line, char **name,
 
 	char *named;
 	int startd,endd,typed;
+	enum labeltype start_labeld, end_labeld;
 
 	/* First, remove any comments on the line. Everything past a ';' is
 	 * ignored */
@@ -293,8 +318,27 @@ int block_load_line(char *line, char **name,
 
 	tok=strtok(NULL, " \t\r\n");
 	
-	if(tok==NULL || strcasecmp(tok, "start")) {
-		msg(0, "Error: 'start' expected\n");
+	if(tok == NULL) {
+		msg(0, "Error: 'unlabeled', 'start' or 'first' expected\n");
+		return -1;
+	}
+
+	if(!strcasecmp(tok, "start")) {
+		start_labeld = LABEL_START_END;
+	} else if(!strcasecmp(tok, "first")) {
+		start_labeld = LABEL_FIRST_LAST;
+	} else if(!strcasecmp(tok, "unlabeled")) {
+		start_labeld = LABEL_NONE;
+
+		tok = strtok(NULL, " \t\r\n");
+		if(tok == NULL || (strcasecmp(tok, "start") &&
+						strcasecmp(tok, "first"))) {
+			msg(0, "Error: 'start' or 'first' expected, "
+						"got '%s' instead\n", tok);
+			return -1;
+		}
+	} else {
+		msg(0, "Error: 'unlabeled', 'start' or 'first' expected\n");
 		return -1;
 	}
 
@@ -313,8 +357,33 @@ int block_load_line(char *line, char **name,
 
 	tok=strtok(NULL, " \t\r\n");
 	
-	if(tok==NULL || strcasecmp(tok, "end")) {
-		msg(0, "Error: 'end' expected\n");
+	if(tok == NULL) {
+		msg(0, "Error: 'unlabeled', 'end' or 'last' expected\n");
+		return -1;
+	}
+
+	int is_last;
+	if(!strcasecmp(tok, "end")) {
+		end_labeld = LABEL_START_END;
+		is_last = 0;
+	} else if(!strcasecmp(tok, "last")) {
+		end_labeld = LABEL_FIRST_LAST;
+		is_last = 1;
+	} else if(!strcasecmp(tok, "unlabeled")) {
+		end_labeld = LABEL_NONE;
+
+		tok = strtok(NULL, " \t\r\n");
+		if(tok != NULL && !strcasecmp(tok, "end")) {
+			is_last = 0;
+		} else if(tok != NULL && !strcasecmp(tok, "last")) {
+			is_last = 1;
+		} else {
+			msg(0, "Error: 'end' or 'last' expected, "
+						"got '%s' instead\n", tok);
+			return -1;
+		}
+	} else {
+		msg(0, "Error: 'unlabeled', 'end' or 'last' expected\n");
 		return -1;
 	}
 
@@ -326,10 +395,20 @@ int block_load_line(char *line, char **name,
 	}
 
 	r=sscanf(tok, "%i", &endd);
-	if(r!=1 || endd <= startd) {
+	if(r != 1) {
 		msg(0, "Error: Invalid end address '%s'\n", tok);
 		return -1;
-	} 
+	}
+
+	if(is_last) {
+		endd++;
+	}
+
+	if(endd <= startd) {
+		msg(0, "Error: End address '%x' before start address '%x'\n",
+								endd, startd);
+		return -1;
+	}
 
 	tok=strtok(NULL, " \t\r\n");
 	
@@ -362,6 +441,8 @@ int block_load_line(char *line, char **name,
 	*start=startd;
 	*end=endd;
 	*type=typed;
+	*start_label = start_labeld;
+	*end_label = end_labeld;
 
 	return 1;
 }
